@@ -23,6 +23,7 @@ const GLib = imports.gi.GLib;
 const GObject = imports.gi.GObject;
 const Gtk = imports.gi.Gtk;
 const Gdk = imports.gi.Gdk;
+const Cairo = imports.cairo;
 const Pango = imports.gi.Pango;
 const PangoCairo = imports.gi.PangoCairo;
 const Gc = imports.gi.Gc;
@@ -53,6 +54,9 @@ const CharacterListRow = new Lang.Class({
         this.parent(params);
         this._characters = filtered.characters;
         this._fontDescription = filtered.fontDescription;
+        var fontDescription = filtered.fontDescription.copy();
+        fontDescription.set_size(fontDescription.get_size() * 0.18);
+        this._overlayFontDescription = fontDescription;
     },
 
     draw: function(cr, x, y, width, height) {
@@ -71,15 +75,118 @@ const CharacterListRow = new Lang.Class({
         // Draw characters.  Do centering and attach to the baseline.
         let cellSize = getCellSize(this._fontDescription);
         for (let i in this._characters) {
-            layout.set_text(this._characters[i], -1);
-            let layoutBaseline = layout.get_baseline() / Pango.SCALE;
-            let [logicalRect, inkRect] = layout.get_extents();
-            cr.moveTo(x + cellSize * i - logicalRect.x / Pango.SCALE +
-                      (cellSize - logicalRect.width / Pango.SCALE) / 2,
-                      y + BASELINE_OFFSET * height - layoutBaseline);
-            PangoCairo.show_layout(cr, layout);
+            var cellRect = new Gdk.Rectangle({ x: x + cellSize * i,
+                                               y: y,
+                                               width: cellSize,
+                                               height: cellSize });
+            if (Gc.character_is_invisible(this._characters[i])) {
+                this._drawBoundingBox(cr, cellRect, this._characters[i]);
+                this._drawCharacterName(cr, cellRect, this._characters[i]);
+            } else {
+                layout.set_text(this._characters[i], -1);
+                if (layout.get_unknown_glyphs_count () == 0) {
+                    let layoutBaseline = layout.get_baseline();
+                    let [logicalRect, inkRect] = layout.get_extents();
+                    cr.moveTo(x + cellSize * i - logicalRect.x / Pango.SCALE +
+                              (cellSize - logicalRect.width / Pango.SCALE) / 2,
+                              y + BASELINE_OFFSET * height -
+                              layoutBaseline / Pango.SCALE);
+                    PangoCairo.show_layout(cr, layout);
+                } else {
+                    this._drawBoundingBox(cr, cellRect, this._characters[i]);
+                    this._drawCharacterName(cr, cellRect, this._characters[i]);
+                }
+            }
         }
     },
+
+    _computeBoundingBox: function(cr, cellRect, uc) {
+        let layout = PangoCairo.create_layout(cr);
+        layout.set_font_description(this._fontDescription);
+        layout.set_text(uc, -1);
+
+        let shapeRect;
+        let layoutBaseline;
+        if (layout.get_unknown_glyphs_count() == 0) {
+            let [logicalRect, inkRect] = layout.get_extents();
+            layoutBaseline = layout.get_baseline();
+            shapeRect = inkRect;
+        } else {
+            // If the character cannot be rendered with the current
+            // font settings, show a rectangle calculated from the
+            // base glyph ('A').
+            if (this._baseGlyphRect == null) {
+                layout.set_text('A', -1);
+                let [baseLogicalRect, baseInkRect] = layout.get_extents();
+                this._baseGlyphLayoutBaseline = layout.get_baseline();
+                this._baseGlyphRect = baseInkRect;
+            }
+            layoutBaseline = this._baseGlyphLayoutBaseline;
+            shapeRect = new Pango.Rectangle({
+                x: this._baseGlyphRect.x,
+                y: this._baseGlyphRect.y,
+                width: this._baseGlyphRect.width,
+                height: this._baseGlyphRect.height
+            });
+            let characterWidth = Gc.character_width (uc);
+            if (characterWidth > 1)
+                shapeRect.width *= characterWidth;
+        }
+
+        shapeRect.x = cellRect.x - shapeRect.x / Pango.SCALE +
+            (cellRect.width - shapeRect.width / Pango.SCALE) / 2;
+        shapeRect.y = cellRect.y + BASELINE_OFFSET * cellRect.height -
+            layoutBaseline / Pango.SCALE;
+        shapeRect.width = shapeRect.width / Pango.SCALE;
+        shapeRect.height = shapeRect.height / Pango.SCALE;
+        return shapeRect;
+    },
+
+    _drawBoundingBox: function(cr, cellRect, uc) {
+        cr.save();
+        cr.rectangle(cellRect.x, cellRect.y, cellRect.width, cellRect.height);
+        cr.clip();
+
+        let layout = PangoCairo.create_layout(cr);
+        layout.set_font_description(this._fontDescription);
+        layout.set_text(uc, -1);
+        let shapeRect = this._computeBoundingBox(cr, cellRect, uc);
+
+        let borderWidth = 1;
+        cr.rectangle(shapeRect.x - borderWidth * 2,
+                     shapeRect.y - borderWidth * 2,
+                     shapeRect.width + borderWidth * 2,
+                     shapeRect.height + borderWidth * 2);
+        cr.setSourceRGBA(239.0 / 255.0, 239.0 / 255.0, 239.0 / 255.0, 1.0);
+        cr.fill();
+
+        cr.restore();
+    },
+
+    _drawCharacterName: function(cr, cellRect, uc) {
+        cr.save();
+        cr.rectangle(cellRect.x, cellRect.y, cellRect.width, cellRect.height);
+        cr.clip();
+
+        let layout = PangoCairo.create_layout(cr);
+        layout.set_width(cellRect.width * Pango.SCALE * 0.8);
+        layout.set_height(cellRect.height * Pango.SCALE * 0.8);
+        layout.set_wrap(Pango.WrapMode.WORD);
+        layout.set_ellipsize(Pango.EllipsizeMode.END);
+        layout.set_alignment(Pango.Alignment.CENTER);
+        layout.set_font_description(this._overlayFontDescription);
+        var name = Gc.character_name(uc);
+        layout.set_text(Util.capitalize(name), -1);
+        let [logicalRect, inkRect] = layout.get_extents();
+        cr.moveTo(cellRect.x - logicalRect.x / Pango.SCALE +
+                  (cellRect.width - logicalRect.width / Pango.SCALE) / 2,
+                  cellRect.y - logicalRect.y / Pango.SCALE +
+                  (cellRect.height - logicalRect.height / Pango.SCALE) / 2);
+        cr.setSourceRGBA(0.0, 0.0, 0.0, 1.0);
+        PangoCairo.show_layout(cr, layout);
+
+        cr.restore();
+    }
 });
 
 const CharacterListWidget = new Lang.Class({
@@ -93,7 +200,9 @@ const CharacterListWidget = new Lang.Class({
         let filtered = Params.filter(params, { fontDescription: null });
         params = Params.fill(params, {});
         this.parent(params);
-        this.get_style_context().add_class('character-list');
+        let context = this.get_style_context();
+        context.add_class('character-list');
+        context.save();
         this._cellsPerRow = CELLS_PER_ROW;
         this._fontDescription = filtered.fontDescription;
         this._characters = [];
@@ -264,13 +373,34 @@ const CharacterListView = new Lang.Class({
             visible: true
         });
         scroll.add(this._characterList);
-        scroll.get_style_context().add_class('character-list-scroll');
+        let context = scroll.get_style_context();
+        context.add_class('character-list-scroll');
+        context.save();
         this.add_named(scroll, 'character-list');
         this.visible_child_name = 'character-list';
 
         this._characters = [];
         this._spinnerTimeoutId = 0;
+        this._searchContext = null;
         this._cancellable = new Gio.Cancellable();
+        this._cancellable.connect(Lang.bind(this, function () {
+            this._stopSpinner();
+            this._searchContext = null;
+            this._characters = [];
+            this.updateCharacterList();
+        }));
+        scroll.connect('edge-reached', Lang.bind(this, this._onEdgeReached));
+    },
+
+    _startSpinner: function() {
+        this._stopSpinner();
+        this._spinnerTimeoutId =
+            GLib.timeout_add(GLib.PRIORITY_DEFAULT, 1000,
+                             Lang.bind(this, function () {
+                                 this._loading_spinner.start();
+                                 this.visible_child_name = 'loading';
+                                 this.show_all();
+                             }));
     },
 
     _stopSpinner: function() {
@@ -279,20 +409,6 @@ const CharacterListView = new Lang.Class({
             this._spinnerTimeoutId = 0;
             this._loading_spinner.stop();
         }
-        this._cancellable.reset();
-    },
-
-    _startSearch: function() {
-        this._cancellable.cancel();
-        this._stopSpinner();
-
-        this._spinnerTimeoutId =
-            GLib.timeout_add(GLib.PRIORITY_DEFAULT, 1000,
-                             Lang.bind(this, function () {
-                                 this._loading_spinner.start();
-                                 this.visible_child_name = 'loading';
-                                 this.show_all();
-                             }));
     },
 
     _finishSearch: function(result) {
@@ -343,63 +459,66 @@ const CharacterListView = new Lang.Class({
         this.show_all();
     },
 
+    _onEdgeReached: function(scrolled, pos) {
+        if (pos == Gtk.PositionType.BOTTOM &&
+            this._searchContext != null && !this._searchContext.is_finished()) {
+            this._searchWithContext(this._searchContext, MAX_SEARCH_RESULTS);
+        }
+    },
+
+    _addSearchResult: function(result) {
+        for (let index = 0; index < result.len; index++) {
+            this._characters.push(Gc.search_result_get(result, index));
+        }
+
+        this.updateCharacterList()
+    },
+
+    _searchWithContext: function(context, count) {
+        this._startSpinner();
+        context.search(
+            count,
+            this._cancellable,
+            Lang.bind(this, function(context, res, user_data) {
+                this._stopSpinner();
+                try {
+                    let result = context.search_finish(res);
+                    this._addSearchResult(result);
+                } catch (e) {
+                    log("Failed to search: " + e);
+                }
+            }));
+    },
+
     searchByCategory: function(category) {
         if ('scripts' in category) {
             this.searchByScripts(category.scripts);
             return;
         }
 
-        this._startSearch()
-        Gc.search_by_category(
-            category.category,
-            -1,
-            this._cancellable,
-            Lang.bind(this,
-                      function(source_object, res, user_data) {
-                          try {
-                              let result = Gc.search_finish(res);
-                              this._finishSearch(result);
-                          } catch (e) {
-                              log("Failed to search by category: " + e);
-                          }
-                      }));
+        this._characters = [];
+        let criteria = Gc.SearchCriteria.new_category(category.category);
+        this._searchContext = new Gc.SearchContext({ criteria: criteria });
+        this._searchWithContext(this._searchContext, MAX_SEARCH_RESULTS);
     },
 
     searchByKeywords: function(keywords) {
-        this._startSearch()
-        Gc.search_by_keywords(
-            keywords,
-            MAX_SEARCH_RESULTS,
-            this._cancellable,
-            Lang.bind(this, function(source_object, res, user_data) {
-                try {
-                    let result = Gc.search_finish(res);
-                    this._finishSearch(result);
-                } catch (e) {
-                    log("Failed to search by keywords: " + e);
-                }
-            }));
+        this._characters = [];
+        let criteria = Gc.SearchCriteria.new_keywords(keywords);
+        this._searchContext = new Gc.SearchContext({ criteria: criteria });
+        this._searchWithContext(this._searchContext, MAX_SEARCH_RESULTS);
     },
 
     searchByScripts: function(scripts) {
-        this._startSearch()
-        Gc.search_by_scripts(
-            scripts,
-            -1,
-            this._cancellable,
-            Lang.bind(this, function(source_object, res, user_data) {
-                try {
-                    let result = Gc.search_finish(res);
-                    this._finishSearch(result);
-                } catch (e) {
-                    log("Failed to search by scripts: " + e);
-                }
-            }));
+        this._characters = [];
+        var criteria = Gc.SearchCriteria.new_scripts(scripts);
+        this._searchContext = new Gc.SearchContext({ criteria: criteria });
+        this._searchWithContext(this._searchContext, MAX_SEARCH_RESULTS);
     },
 
     cancelSearch: function() {
         this._cancellable.cancel();
-        this._finishSearch([]);
+        this._cancellable.reset();
     },
 
     setFilterFont: function(family) {
